@@ -1,21 +1,24 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 from itertools import chain
+
 try:
     import phonenumbers
-    from phonenumbers import PhoneNumberFormat, NumberParseException
+    from phonenumbers import NumberParseException, PhoneNumberFormat
 except ImportError:
     phonenumbers = None
 
 from trytond.i18n import gettext
 from trytond.model import (
-    ModelView, ModelSQL, DeactivableMixin, MultiValueMixin, ValueMixin, fields,
-    sequence_ordered)
+    DeactivableMixin, Index, ModelSQL, ModelView, MultiValueMixin, ValueMixin,
+    fields, sequence_ordered)
 from trytond.model.exceptions import AccessError
 from trytond.pyson import Eval
+from trytond.tools.email_ import (
+    EmailNotValidError, normalize_email, validate_email)
 from trytond.transaction import Transaction
 
-from .exceptions import InvalidPhoneNumber
+from .exceptions import InvalidEMail, InvalidPhoneNumber
 
 _TYPES = [
     ('phone', 'Phone'),
@@ -37,6 +40,31 @@ _PHONE_TYPES = {
     }
 
 
+class _ContactMechanismMixin:
+    __slots__ = ()
+
+    def contact_mechanism_get(self, types=None, usage=None):
+        """
+        Try to find a contact mechanism for the given types and usage, if no
+        usage matches the first mechanism of the given types is returned.
+        """
+        default_mechanism = None
+        if types:
+            if isinstance(types, str):
+                types = {types}
+            mechanisms = [m for m in self.contact_mechanisms
+                if m.type in types]
+        else:
+            mechanisms = self.contact_mechanisms
+        if mechanisms:
+            default_mechanism = mechanisms[0]
+            if usage:
+                for mechanism in mechanisms:
+                    if getattr(mechanism, usage):
+                        return mechanism
+        return default_mechanism
+
+
 class ContactMechanism(
         DeactivableMixin, sequence_ordered(), ModelSQL, ModelView,
         MultiValueMixin):
@@ -45,16 +73,22 @@ class ContactMechanism(
     _rec_name = 'value'
 
     type = fields.Selection(_TYPES, "Type", required=True, sort=False)
-    value = fields.Char("Value", select=True,
+    value = fields.Char(
+        "Value",
         # Add all function fields to ensure to always fill them via on_change
-        depends=[
+        depends={
             'email', 'website', 'skype', 'sip', 'other_value',
-            'value_compact'])
+            'value_compact'})
     value_compact = fields.Char('Value Compact', readonly=True)
     name = fields.Char("Name")
     comment = fields.Text("Comment")
     party = fields.Many2One(
-        'party.party', "Party", required=True, ondelete='CASCADE', select=True)
+        'party.party', "Party", required=True, ondelete='CASCADE')
+    address = fields.Many2One(
+        'party.address', "Address", ondelete='CASCADE',
+        domain=[
+            ('party', '=', Eval('party', -1)),
+            ])
     language = fields.MultiValue(
         fields.Many2One('ir.lang', "Language",
             help="Used to translate communication made "
@@ -65,27 +99,27 @@ class ContactMechanism(
     email = fields.Function(fields.Char('E-Mail', states={
         'invisible': Eval('type') != 'email',
         'required': Eval('type') == 'email',
-        }, depends=['value', 'type']),
+        }, depends={'value'}),
         'get_value', setter='set_value')
     website = fields.Function(fields.Char('Website', states={
         'invisible': Eval('type') != 'website',
         'required': Eval('type') == 'website',
-        }, depends=['value', 'type']),
+        }, depends={'value'}),
         'get_value', setter='set_value')
     skype = fields.Function(fields.Char('Skype', states={
         'invisible': Eval('type') != 'skype',
         'required': Eval('type') == 'skype',
-        }, depends=['value', 'type']),
+        }, depends={'value'}),
         'get_value', setter='set_value')
     sip = fields.Function(fields.Char('SIP', states={
         'invisible': Eval('type') != 'sip',
         'required': Eval('type') == 'sip',
-        }, depends=['value', 'type']),
+        }, depends={'value'}),
         'get_value', setter='set_value')
     other_value = fields.Function(fields.Char('Value', states={
         'invisible': Eval('type').in_(['email', 'website', 'skype', 'sip']),
         'required': ~Eval('type').in_(['email', 'website']),
-        }, depends=['value', 'type']),
+        }, depends={'value'}),
         'get_value', setter='set_value')
     url = fields.Function(fields.Char('URL', states={
                 'invisible': ~Eval('url'),
@@ -94,7 +128,12 @@ class ContactMechanism(
 
     @classmethod
     def __setup__(cls):
+        cls.value.search_unaccented = False
+        cls.value_compact.search_unaccented = False
         super(ContactMechanism, cls).__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.add(
+            Index(t, (Index.Unaccent(t.value_compact), Index.Similarity())))
         cls._order.insert(0, ('party.distance', 'ASC NULLS LAST'))
         cls._order.insert(1, ('party', 'ASC'))
 
@@ -106,6 +145,11 @@ class ContactMechanism(
     def default_party(cls):
         return Transaction().context.get('related_party')
 
+    @fields.depends('address', '_parent_address.party')
+    def on_change_address(self):
+        if self.address:
+            self.party = self.address.party
+
     @classmethod
     def get_value(cls, mechanisms, names):
         return dict((name, dict((m.id, m.value) for m in mechanisms))
@@ -141,9 +185,12 @@ class ContactMechanism(
         for country_code in chain(self._phone_country_codes(), [None]):
             try:
                 # Country code is ignored if value has an international prefix
-                return phonenumbers.parse(value, country_code)
+                phonenumber = phonenumbers.parse(value, country_code)
             except NumberParseException:
                 pass
+            else:
+                if phonenumbers.is_valid_number(phonenumber):
+                    return phonenumber
         return None
 
     @fields.depends(methods=['_parse_phonenumber'])
@@ -153,6 +200,8 @@ class ContactMechanism(
             if phonenumber:
                 value = phonenumbers.format_number(
                     phonenumber, PhoneNumberFormat.INTERNATIONAL)
+        if type_ == 'email' and value:
+            value = normalize_email(value)
         return value
 
     @fields.depends(methods=['_parse_phonenumber'])
@@ -258,19 +307,39 @@ class ContactMechanism(
         cls._format_values(all_mechanisms)
 
     @classmethod
-    def validate(cls, mechanisms):
-        super(ContactMechanism, cls).validate(mechanisms)
-        for mechanism in mechanisms:
-            mechanism.check_valid_phonenumber()
+    def validate_fields(cls, mechanisms, field_names):
+        super().validate_fields(mechanisms, field_names)
+        cls.check_valid_phonenumber(mechanisms, field_names)
+        cls.check_valid_email(mechanisms, field_names)
 
-    def check_valid_phonenumber(self):
-        if not phonenumbers or self.type not in _PHONE_TYPES:
+    @classmethod
+    def check_valid_phonenumber(cls, mechanisms, field_names=None):
+        if field_names and not (field_names & {'type', 'value'}):
             return
-        phonenumber = self._parse_phonenumber(self.value)
-        if not (phonenumber and phonenumbers.is_valid_number(phonenumber)):
-            raise InvalidPhoneNumber(
-                gettext('party.msg_invalid_phone_number',
-                    phone=self.value, party=self.party.rec_name))
+        if not phonenumbers:
+            return
+        for mechanism in mechanisms:
+            if mechanism.type not in _PHONE_TYPES:
+                continue
+            if not mechanism._parse_phonenumber(mechanism.value):
+                raise InvalidPhoneNumber(
+                    gettext('party.msg_invalid_phone_number',
+                        phone=mechanism.value, party=mechanism.party.rec_name))
+
+    @classmethod
+    def check_valid_email(cls, mechanisms, field_names=None):
+        if field_names and not (field_names & {'type', 'value'}):
+            return
+        for mechanism in mechanisms:
+            if mechanism.type == 'email' and mechanism.value:
+                try:
+                    validate_email(mechanism.value)
+                except EmailNotValidError as e:
+                    raise InvalidEMail(gettext(
+                            'party.msg_email_invalid',
+                            email=mechanism.value,
+                            party=mechanism.party.rec_name),
+                        str(e)) from e
 
     @classmethod
     def usages(cls, _fields=None):
@@ -287,5 +356,5 @@ class ContactMechanismLanguage(ModelSQL, ValueMixin):
     __name__ = 'party.contact_mechanism.language'
     contact_mechanism = fields.Many2One(
         'party.contact_mechanism', "Contact Mechanism",
-        ondelete='CASCADE', select=True)
+        ondelete='CASCADE')
     language = fields.Many2One('ir.lang', "Language")
