@@ -1,37 +1,51 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-from stdnum import get_cc_module
+
+import logging
+
 import stdnum.exceptions
-from sql import Null, Column, Literal
+from sql import Column, Literal
 from sql.aggregate import Min
-from sql.functions import CharLength, Substring, Position
+from sql.functions import CharLength
+from stdnum import get_cc_module
+from stdnum.eu.vat import MEMBER_STATES as EU_MEMBER_STATES
 
 from trytond.i18n import gettext
-from trytond.model import (ModelView, ModelSQL, MultiValueMixin, ValueMixin,
-    DeactivableMixin, fields, Unique, sequence_ordered)
-from trytond.wizard import Wizard, StateTransition, StateView, Button
-from trytond.pyson import Eval, Bool
-from trytond.transaction import Transaction
+from trytond.model import (
+    DeactivableMixin, Index, ModelSQL, ModelView, MultiValueMixin, Unique,
+    ValueMixin, convert_from, fields, sequence_ordered)
+from trytond.model.exceptions import AccessError
 from trytond.pool import Pool
-from trytond import backend
-from trytond.tools.multivalue import migrate_property
-from trytond.tools import lstrip_wildcard
+from trytond.pyson import Bool, Eval
+from trytond.tools import is_full_text, lstrip_wildcard
+from trytond.transaction import Transaction, inactive_records
+from trytond.wizard import Button, StateTransition, StateView, Wizard
+
+from .contact_mechanism import _PHONE_TYPES, _ContactMechanismMixin
 from .exceptions import (
-    InvalidIdentifierCode, VIESUnavailable, SimilarityWarning, EraseError)
+    EraseError, InvalidIdentifierCode, SimilarityWarning, VIESUnavailable)
+
+logger = logging.getLogger(__name__)
 
 
-class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
+class Party(
+        DeactivableMixin, _ContactMechanismMixin, ModelSQL, ModelView,
+        MultiValueMixin):
     "Party"
     __name__ = 'party.party'
 
+    _contact_mechanism_states = {
+        'readonly': Eval('id', -1) >= 0,
+        }
+
     name = fields.Char(
-        "Name", select=True,
+        "Name", strip=False,
         help="The main identifier of the party.")
-    code = fields.Char('Code', required=True, select=True,
+    code = fields.Char(
+        "Code", required=True,
         states={
             'readonly': Eval('code_readonly', True),
             },
-        depends=['code_readonly'],
         help="The unique identifier of the party.")
     code_readonly = fields.Function(fields.Boolean('Code Readonly'),
         'get_code_readonly')
@@ -59,35 +73,42 @@ class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
             },
         help="The party replacing this one.")
     full_name = fields.Function(fields.Char('Full Name'), 'get_full_name')
-    phone = fields.Function(fields.Char('Phone'), 'get_mechanism')
-    mobile = fields.Function(fields.Char('Mobile'), 'get_mechanism')
-    fax = fields.Function(fields.Char('Fax'), 'get_mechanism')
-    email = fields.Function(fields.Char('E-Mail'), 'get_mechanism')
-    website = fields.Function(fields.Char('Website'), 'get_mechanism')
+    phone = fields.Function(
+        fields.Char("Phone", states=_contact_mechanism_states),
+        'get_contact_mechanism', setter='set_contact_mechanism')
+    mobile = fields.Function(
+        fields.Char("Mobile", states=_contact_mechanism_states),
+        'get_contact_mechanism', setter='set_contact_mechanism')
+    fax = fields.Function(
+        fields.Char("Fax", states=_contact_mechanism_states),
+        'get_contact_mechanism', setter='set_contact_mechanism')
+    email = fields.Function(
+        fields.Char("E-Mail", states=_contact_mechanism_states),
+        'get_contact_mechanism', setter='set_contact_mechanism')
+    website = fields.Function(
+        fields.Char("Website", states=_contact_mechanism_states),
+        'get_contact_mechanism', setter='set_contact_mechanism')
     distance = fields.Function(fields.Integer('Distance'), 'get_distance')
 
+    del _contact_mechanism_states
+
     @classmethod
     def __setup__(cls):
+        cls.code.search_unaccented = False
         super(Party, cls).__setup__()
         t = cls.__table__()
         cls._sql_constraints = [
             ('code_uniq', Unique(t, t.code), 'party.msg_party_code_unique')
             ]
+        cls._sql_indexes.update({
+                Index(t, (t.code, Index.Equality())),
+                Index(t, (t.code, Index.Similarity())),
+                })
         cls._order.insert(0, ('distance', 'ASC NULLS LAST'))
         cls._order.insert(1, ('name', 'ASC'))
         cls.active.states.update({
                 'readonly': Bool(Eval('replaced_by')),
                 })
-        cls.active.depends.append('replaced_by')
-
-    @classmethod
-    def __register__(cls, module_name):
-        super(Party, cls).__register__(module_name)
-
-        table_h = cls.__table_handler__(module_name)
-
-        # Migration from 3.8
-        table_h.not_null_action('name', 'remove')
 
     @staticmethod
     def order_code(tables):
@@ -122,18 +143,7 @@ class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
 
     @classmethod
     def tax_identifier_types(cls):
-        return [
-            'ad_nrt', 'al_nipt', 'ar_cuit', 'be_vat', 'bg_vat', 'by_unp',
-            'ch_vat', 'cl_rut', 'cn_uscc', 'co_rut', 'cr_cpj', 'cz_dic',
-            'de_vat', 'dk_cvr', 'do_rnc', 'ec_ruc', 'ee_kmkr', 'es_nif',
-            'eu_vat', 'fi_alv', 'fr_tva', 'gb_vat', 'gr_vat', 'gt_nit',
-            'hu_anum', 'id_npwp', 'ie_vat', 'il_hp', 'is_vsk', 'it_iva',
-            'jp_cn', 'kr_brn', 'lt_pvm', 'lu_tva', 'lv_pvn', 'mc_tva',
-            'md_idno', 'mt_vat', 'mx_rfc', 'nl_btw', 'no_mva', 'nz_ird',
-            'pe_ruc', 'pl_nip', 'pt_nif', 'py_ruc', 'ro_cf', 'rs_pib',
-            'ru_inn', 'se_vat', 'si_ddv', 'sk_dph', 'sm_coe', 'us_atin',
-            'us_ein', 'us_itin', 'us_ptin', 'us_ssn', 'us_tin', 'uy_ruc',
-            've_rif', 'za_tin']
+        return TAX_IDENTIFIER_TYPES
 
     def get_tax_identifier(self, name):
         types = self.tax_identifier_types()
@@ -144,10 +154,11 @@ class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
     @classmethod
     def search_tax_identifier(cls, name, clause):
         _, operator, value = clause
+        nested = clause[0][len(name) + 1:]
         types = cls.tax_identifier_types()
         domain = [
             ('identifiers', 'where', [
-                    ('code', operator, value),
+                    (nested or 'rec_name', operator, value),
                     ('type', 'in', types),
                     ]),
             ]
@@ -166,12 +177,31 @@ class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
     def get_full_name(self, name):
         return self.name
 
-    def get_mechanism(self, name):
+    def get_contact_mechanism(self, name):
         for mechanism in self.contact_mechanisms:
             if mechanism.type == name:
                 return mechanism.value
         return ''
 
+    @classmethod
+    def set_contact_mechanism(cls, parties, name, value):
+        pool = Pool()
+        ContactMechanism = pool.get('party.contact_mechanism')
+        contact_mechanisms = []
+        for party in parties:
+            if getattr(party, name):
+                type_string = cls.fields_get([name])[name]['string']
+                raise AccessError(gettext(
+                        'party.msg_party_set_contact_mechanism',
+                        party=party.rec_name,
+                        field=type_string))
+            if value:
+                contact_mechanisms.append(ContactMechanism(
+                        party=party,
+                        type=name,
+                        value=value))
+        ContactMechanism.save(contact_mechanisms)
+
     @classmethod
     def _distance_query(cls, usages=None, party=None, depth=None):
         context = Transaction().context
@@ -221,6 +251,14 @@ class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
             query, _ = tables[key][None]
         return [query.distance]
 
+    @classmethod
+    def index_set_field(cls, name):
+        index = super().index_set_field(name)
+        if name in _PHONE_TYPES:
+            # Phone validation may need the address country
+            index = cls.index_set_field('addresses') + 1
+        return index
+
     @classmethod
     def _new_code(cls, **pattern):
         pool = Pool()
@@ -261,18 +299,19 @@ class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
 
     @classmethod
     def search_rec_name(cls, name, clause):
-        if clause[1].startswith('!') or clause[1].startswith('not '):
+        _, operator, operand, *extra = clause
+        if operator.startswith('!') or operator.startswith('not '):
             bool_op = 'AND'
         else:
             bool_op = 'OR'
-        code_value = clause[2]
-        if clause[1].endswith('like'):
-            code_value = lstrip_wildcard(clause[2])
+        code_value = operand
+        if operator.endswith('like') and is_full_text(operand):
+            code_value = lstrip_wildcard(operand)
         return [bool_op,
-            ('code', clause[1], code_value) + tuple(clause[3:]),
-            ('identifiers.code', clause[1], code_value) + tuple(clause[3:]),
-            ('name',) + tuple(clause[1:]),
-            ('contact_mechanisms.rec_name',) + tuple(clause[1:]),
+            ('code', operator, code_value, *extra),
+            ('identifiers.code', operator, code_value, *extra),
+            ('name', operator, operand, *extra),
+            ('contact_mechanisms.rec_name', operator, operand, *extra),
             ]
 
     def address_get(self, type=None):
@@ -289,74 +328,177 @@ class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
                         return address
         return default_address
 
-    def contact_mechanism_get(self, types=None, usage=None):
-        """
-        Try to find a contact mechanism for the given types and usage, if no
-        usage matches the first mechanism of the given types is returned.
-        """
-        default_mechanism = None
-        if types:
-            if isinstance(types, str):
-                types = {types}
-            mechanisms = [m for m in self.contact_mechanisms
-                if m.type in types]
+    @classmethod
+    def autocomplete(cls, text, domain=None, limit=None, order=None):
+        pool = Pool()
+        Configuration = pool.get('party.configuration')
+        configuration = Configuration(1)
+
+        result = super().autocomplete(
+            text, domain=domain, limit=limit, order=order)
+
+        identifier_types = dict(configuration.get_identifier_types())
+
+        def add(type, code):
+            result.append({
+                    'id': None,
+                    'name': code,
+                    'defaults': {
+                        'identifiers': [{
+                                'type': type,
+                                'code': code,
+                                }],
+                        },
+                    })
+
+        eu_vat = get_cc_module('eu', 'vat')
+        if 'eu_vat' in identifier_types and eu_vat.is_valid(text):
+            code = eu_vat.compact(text)
+            add('eu_vat', code)
         else:
-            mechanisms = self.contact_mechanisms
-        if mechanisms:
-            default_mechanism = mechanisms[0]
-            if usage:
-                for mechanism in mechanisms:
-                    if getattr(mechanism, usage):
-                        return mechanism
-        return default_mechanism
+            for country in eu_vat.guess_country(text):
+                if 'eu_vat' in identifier_types:
+                    code = eu_vat.compact(country + text)
+                    add('eu_vat', code)
+                elif f'{country}_vat' in identifier_types:
+                    vat = get_cc_module(country, 'vat')
+                    code = vat.compact(text)
+                    add(f'{country}_vat', code)
+        return result
+
+    @classmethod
+    def default_get(cls, fields_names, with_rec_name=True):
+        pool = Pool()
+        Country = pool.get('country.country')
+
+        values = super().default_get(fields_names, with_rec_name=with_rec_name)
+        eu_vat = get_cc_module('eu', 'vat')
+        eu_types = {'eu_vat'} | {f'{m}_vat' for m in EU_MEMBER_STATES}
+        for identifier in values.get('identifiers') or []:
+            if identifier.get('type') in eu_types and identifier.get('code'):
+                code = identifier['code']
+                if identifier['type'] != 'eu_vat':
+                    code = identifier['type'][2:] + code
+                try:
+                    result = eu_vat.check_vies(code)
+                except Exception:
+                    logger.debug(
+                        f"Fail to check {identifier['code']}", exc_info=True)
+                    continue
+                if result['valid']:
+                    values['name'] = result['name']
+                    for address in values.get('addresses') or []:
+                        if not address:
+                            break
+                    else:
+                        address = {}
+                        values.setdefault('addresses', []).insert(0, address)
+                    address['street'] = result['address']
+                    country_code = code[:2]
+                    countries = Country.search([
+                            ('code', 'ilike', country_code),
+                            ], limit=1)
+                    if countries:
+                        country, = countries
+                        address['country'] = country.id
+                        if with_rec_name:
+                            address.setdefault('country.', {})['rec_name'] = (
+                                country.rec_name)
+                    break
+        return values
 
 
 class PartyLang(ModelSQL, ValueMixin):
     "Party Lang"
     __name__ = 'party.party.lang'
     party = fields.Many2One(
-        'party.party', "Party", ondelete='CASCADE', select=True)
+        'party.party', "Party", ondelete='CASCADE')
     lang = fields.Many2One('ir.lang', "Language")
 
-    @classmethod
-    def __register__(cls, module_name):
-        pool = Pool()
-        Party = pool.get('party.party')
-        cursor = Transaction().connection.cursor()
-        exist = backend.TableHandler.table_exist(cls._table)
-        table = cls.__table__()
-        party = Party.__table__()
-
-        super(PartyLang, cls).__register__(module_name)
-
-        if not exist:
-            party_h = Party.__table_handler__(module_name)
-            if party_h.column_exist('lang'):
-                query = table.insert(
-                    [table.party, table.lang],
-                    party.select(party.id, party.lang))
-                cursor.execute(*query)
-                party_h.drop_column('lang')
-            else:
-                cls._migrate_property([], [], [])
-
-    @classmethod
-    def _migrate_property(cls, field_names, value_names, fields):
-        field_names.append('lang')
-        value_names.append('lang')
-        migrate_property(
-            'party.party', field_names, cls, value_names,
-            parent='party', fields=fields)
-
 
 class PartyCategory(ModelSQL):
     'Party - Category'
     __name__ = 'party.party-party.category'
     _table = 'party_category_rel'
-    party = fields.Many2One('party.party', 'Party', ondelete='CASCADE',
-            required=True, select=True)
-    category = fields.Many2One('party.category', 'Category',
-        ondelete='CASCADE', required=True, select=True)
+    party = fields.Many2One(
+        'party.party', "Party", ondelete='CASCADE', required=True)
+    category = fields.Many2One(
+        'party.category', "Category", ondelete='CASCADE', required=True)
+
+
+IDENTIFIER_VAT = [
+    'ad_nrt',
+    'al_nipt',
+    'ar_cuit',
+    'at_uid',
+    'au_abn',
+    'br_cnpj',
+    'by_unp',
+    'ca_bn',
+    'cl_rut',
+    'cn_uscc',
+    'co_nit',
+    'cr_cpj',
+    'cz_dic',
+    'dk_cvr',
+    'do_rnc',
+    'dz_nif',
+    'ec_ruc',
+    'ee_kmkr',
+    'eg_tn',
+    'es_nif',
+    'fi_alv',
+    'fo_vn',
+    'fr_tva',
+    'gh_tin',
+    'gn_nifp',
+    'gt_nit',
+    'hr_oib',
+    'hu_anum',
+    'id_npwp',
+    'il_hp',
+    'in_gstin',
+    'is_vsk',
+    'it_iva',
+    'jp_cn',
+    'ke_pin',
+    'kr_brn',
+    'lt_pvm',
+    'lu_tva',
+    'lv_pvn',
+    'ma_ice',
+    'mc_tva',
+    'me_pib',
+    'mk_edb',
+    'mx_rfc',
+    'nl_btw',
+    'no_mva',
+    'nz_ird',
+    'pe_ruc',
+    'pl_nip',
+    'pt_nif',
+    'py_ruc',
+    'ro_cf',
+    'rs_pib',
+    'ru_inn',
+    'sg_uen',
+    'si_ddv',
+    'sk_dph',
+    'sm_coe',
+    'sv_nit',
+    'tn_mf',
+    'tr_vkn',
+    'tw_ubn',
+    'uy_rut',
+    've_rif',
+    'vn_mst',
+    ]
+
+
+def replace_vat(code):
+    if code in IDENTIFIER_VAT:
+        code = code.split('_', 1)[0] + '_vat'
+    return code
 
 
 IDENTIFIER_TYPES = [
@@ -366,10 +508,13 @@ IDENTIFIER_TYPES = [
     ('ar_dni', "Argentinian National Identity Number"),
     ('at_businessid', "Austrian Company Register"),
     ('at_tin', "Austrian Tax Identification"),
+    ('at_uid', "Austrian Umsatzsteuer-Identifikationsnummer"),
     ('at_vnr', "Austrian Social Security Number"),
     ('au_abn', "Australian Business Number"),
     ('au_acn', "Australian Company Number"),
     ('au_tfn', "Australian Tax File Number"),
+    ('be_bis', "Belgian BIS Number"),
+    ('be_nn', "Belgian National Number"),
     ('be_vat', "Belgian Enterprise Number"),
     ('bg_egn', "Bulgarian Personal Identity Codes"),
     ('bg_pnf', "Bulgarian Number of a Foreigner"),
@@ -402,31 +547,37 @@ IDENTIFIER_TYPES = [
     ('dk_cvr', "Danish VAT Number"),
     ('do_cedula', "Dominican Republic National Identification Number"),
     ('do_rnc', "Dominican Republic Tax"),
+    ('dz_nif', "Algerian Tax Number"),
     ('ec_ci', "Ecuadorian Personal Identity Code"),
     ('ec_ruc', "Ecuadorian Tax Identification"),
     ('ee_ik', "Estonian Personal ID Number"),
     ('ee_kmkr', "Estonian VAT Number"),
     ('ee_registrikood', "Estonian Organisation Registration Code"),
+    ('eg_tn', "Egyptian Tax Registration Number"),
     ('es_cif', "Spanish Company Tax"),
     ('es_dni', "Spanish Personal Identity Codes"),
     ('es_nie', "Spanish Foreigner Number"),
     ('es_nif', "Spanish VAT Number"),
     ('eu_at_02', "SEPA Identifier of the Creditor (AT-02)"),
+    ('eu_oss', "European VAT on e-Commerce - One Stop Shop"),
     ('eu_vat', "European VAT Number"),
     ('fi_alv', "Finnish VAT Number"),
     ('fi_associationid', "Finnish Association Identifier"),
     ('fi_hetu', "Finnish Personal Identity Code"),
     ('fi_veronumero', "Finnish Individual Tax Number"),
     ('fi_ytunnus', "Finnish Business Identifier"),
+    ('fo_vn', "Faroese Tax Number"),
     ('fr_nif', "French Tax Identification Number"),
     ('fr_nir', "French Personal Identification Number"),
-    # TODO: remove from party_siren
-    # ('fr_siren', "French Company Identification Number"),
+    ('fr_siren', "French Company Identification Number"),
+    ('fr_siret', "French Company Establishment Identification Number"),
     ('fr_tva', "French VAT Number"),
     ('gb_nhs',
         "United Kingdom National Health Service Patient Identifier"),
     ('gb_upn', "English Unique Pupil Number"),
     ('gb_vat', "United Kingdom (and Isle of Man) VAT Number"),
+    ('gh_tin', "Ghanaian Taxpayer Identification Number"),
+    ('gn_nifp', "Guinean Tax Number"),
     ('gr_amka', "Greek Social Security Number"),
     ('gr_vat', "Greek VAT Number"),
     ('gt_nit', "Guatemala Tax Number"),
@@ -438,6 +589,8 @@ IDENTIFIER_TYPES = [
     ('il_hp', "Israeli Company Number"),
     ('il_idnr', "Israeli Identity Number"),
     ('in_aadhaar', "Indian Digital Resident Personal Identity Number"),
+    ('in_epic', "Indian Voter ID"),
+    ('in_gstin', "Indian VAT number"),
     ('in_pan', "Indian Income Tax Identifier"),
     ('is_kennitala',
         "Icelandic Personal and Organisation Identity Code"),
@@ -451,8 +604,10 @@ IDENTIFIER_TYPES = [
     ('lt_pvm', "Lithuanian VAT Number"),
     ('lu_tva', "Luxembourgian VAT Number"),
     ('lv_pvn', "Latvian VAT Number"),
+    ('ma_ice', "Moroccan Tax Number"),
     ('mc_tva', "Monacan VAT Number"),
     ('md_idno', "Moldavian Company Identification Number"),
+    ('mk_edb', "Macedonian Tax Number"),
     ('mt_vat', "Maltese VAT Number"),
     ('mu_nid', "Mauritian National Identifier"),
     ('mx_rfc', "Mexican Tax Number"),
@@ -469,9 +624,11 @@ IDENTIFIER_TYPES = [
     ('nz_ird', "New Zealand Inland Revenue Department Number"),
     ('pe_cui', "Peruvian Identity Number"),
     ('pe_ruc', "Peruvian Company Tax Number"),
+    ('pk_cnic', "Pakistani Computerised National Identity Card Number"),
     ('pl_nip', "Polish VAT Number"),
     ('pl_pesel', "Polish National Identification Number"),
     ('pl_regon', "Polish Register of Economic Units"),
+    ('pt_cc', "Portuguese Identity Number"),
     ('pt_nif', "Portuguese VAT Number"),
     ('py_ruc', "Paraguay Tax Number"),
     ('ro_cf', "Romanian VAT Number"),
@@ -482,11 +639,19 @@ IDENTIFIER_TYPES = [
     ('se_orgnr', "Swedish Company Number"),
     ('se_personnummer', "Swedish Personal Number"),
     ('se_vat', "Swedish VAT Number"),
+    ('si_businessid', "Slovenian Corporate Registration Number"),
     ('si_ddv', "Slovenian VAT Number"),
+    ('si_emso', "Slovenian Unique Master Citizen Number"),
     ('sk_dph', "Slovak VAT Number"),
     ('sk_rc', "Slovak Birth Number"),
     ('sm_coe', "San Marino National Tax Number"),
+    ('th_moa', "Thai Memorandum of Association Number"),
+    ('th_pin', "Thai Personal Identification Number"),
+    ('th_tin', "Thai Taxpayer Identification Number"),
+    ('tn_mf', "Tunisian Tax Number"),
     ('tr_tckimlik', "Turkish Personal Identification Number"),
+    ('ua_edrpou', "Ukrainian Identifier for Enterprises and Organizations"),
+    ('ua_rntrc', "Ukrainian Individual Taxpayer Registration Number"),
     ('us_atin', "U.S. Adoption Taxpayer Identification Number"),
     ('us_ein', "U.S. Employer Identification Number"),
     ('us_itin', "U.S. Individual Taxpayer Identification Number"),
@@ -495,60 +660,140 @@ IDENTIFIER_TYPES = [
     ('us_tin', "U.S. Taxpayer Identification Number"),
     ('uy_ruc', "Uruguay Tax Number"),
     ('ve_rif', "Venezuelan VAT Number"),
+    ('vn_mst', "Vietnam Tax Number"),
     ('za_idnr', "South African Identity Document Number"),
     ('za_tin', "South African Tax Identification Number"),
     ]
 
+IDENTIFIER_TYPES = list(map(
+        lambda x: (replace_vat(x[0]), x[1]),
+        IDENTIFIER_TYPES))
 
-class Identifier(sequence_ordered(), ModelSQL, ModelView):
+TAX_IDENTIFIER_TYPES = [
+    'ad_nrt',
+    'al_nipt',
+    'ar_cuit',
+    'at_uid',
+    'au_abn',
+    'au_acn',
+    'be_vat',
+    'bg_vat',
+    'by_unp',
+    'ch_vat',
+    'cl_rut',
+    'cn_uscc',
+    'co_rut',
+    'cr_cpj',
+    'cz_dic',
+    'de_vat',
+    'dk_cvr',
+    'do_rnc',
+    'dz_nif',
+    'ec_ruc',
+    'ee_kmkr',
+    'eg_tn',
+    'es_nif',
+    'eu_vat',
+    'fi_alv',
+    'fo_vn',
+    'fr_tva',
+    'gb_vat',
+    'gh_tin',
+    'gn_nifp',
+    'gr_vat',
+    'gt_nit',
+    'hu_anum',
+    'id_npwp',
+    'ie_vat',
+    'il_hp',
+    'in_gstin',
+    'is_vsk',
+    'it_iva',
+    'jp_cn',
+    'kr_brn',
+    'lt_pvm',
+    'lu_tva',
+    'lv_pvn',
+    'ma_ice',
+    'mc_tva',
+    'md_idno',
+    'mk_edb',
+    'mt_vat',
+    'mx_rfc',
+    'nl_btw',
+    'no_mva',
+    'nz_ird',
+    'pe_ruc',
+    'pl_nip',
+    'pt_nif',
+    'py_ruc',
+    'ro_cf',
+    'rs_pib',
+    'ru_inn',
+    'se_vat',
+    'si_ddv',
+    'sk_dph',
+    'sm_coe',
+    'th_tin',
+    'tn_mf',
+    'ua_edrpou',
+    'ua_rntrc',
+    'us_atin',
+    'us_ein',
+    'us_itin',
+    'us_ptin',
+    'us_ssn',
+    'us_tin',
+    'uy_ruc',
+    've_rif',
+    'vn_mst',
+    'za_tin',
+    ]
+
+TAX_IDENTIFIER_TYPES = list(map(replace_vat, TAX_IDENTIFIER_TYPES))
+
+
+class Identifier(sequence_ordered(), DeactivableMixin, ModelSQL, ModelView):
     'Party Identifier'
     __name__ = 'party.identifier'
     _rec_name = 'code'
-    party = fields.Many2One('party.party', 'Party', ondelete='CASCADE',
-        required=True, select=True,
+    party = fields.Many2One(
+        'party.party', "Party", ondelete='CASCADE', required=True,
         help="The party identified by this record.")
+    address = fields.Many2One(
+        'party.address', "Address", ondelete='CASCADE',
+        states={
+            'required': Eval('type_address', False),
+            'invisible': ~Eval('type_address', True),
+            },
+        domain=[
+            ('party', '=', Eval('party', -1)),
+            ],
+        help="The address identified by this record.")
     type = fields.Selection('get_types', 'Type')
     type_string = type.translated('type')
+    type_address = fields.Function(
+        fields.Boolean("Type of Address"), 'on_change_with_type_address')
     code = fields.Char('Code', required=True)
 
     @classmethod
     def __register__(cls, module_name):
-        pool = Pool()
-        Party = pool.get('party.party')
         cursor = Transaction().connection.cursor()
-        party = Party.__table__()
         table = cls.__table__()
 
         super().__register__(module_name)
 
-        party_h = Party.__table_handler__(module_name)
-        if (party_h.column_exist('vat_number')
-                and party_h.column_exist('vat_country')):
-            identifiers = []
-            cursor.execute(*party.select(
-                    party.id, party.vat_number, party.vat_country,
-                    where=(party.vat_number != Null)
-                    | (party.vat_country != Null)))
-            for party_id, number, country in cursor:
-                code = (country or '') + (number or '')
-                if not code:
-                    continue
-                for type in Party.tax_identifier_types():
-                    module = get_cc_module(*type.split('_', 1))
-                    if module.is_valid(code):
-                        break
-                else:
-                    type = None
-                identifiers.append(
-                    cls(party=party_id, code=code, type=type))
-            cls.save(identifiers)
-            party_h.drop_column('vat_number')
-            party_h.drop_column('vat_country')
-
         # Migration from 5.8: Rename cn_rit into cn_ric
         cursor.execute(*table.update([table.type], ['cn_ric'],
                 where=(table.type == 'cn_rit')))
 
+        # Migration from 6.8: Use vat alias
+        for old in IDENTIFIER_VAT:
+            new = replace_vat(old)
+            cursor.execute(*table.update(
+                    [table.type], [new],
+                    where=table.type == old))
+
     @classmethod
     def get_types(cls):
         pool = Pool()
@@ -556,6 +801,19 @@ class Identifier(sequence_ordered(), ModelSQL, ModelView):
         configuration = Configuration(1)
         return [(None, '')] + configuration.get_identifier_types()
 
+    @classmethod
+    def _type_addresses(cls):
+        return {'fr_siret'}
+
+    @fields.depends('address', '_parent_address.party')
+    def on_change_address(self):
+        if self.address:
+            self.party = self.address.party
+
+    @fields.depends('type')
+    def on_change_with_type_address(self, name=None):
+        return self.type in self._type_addresses()
+
     @fields.depends('type', 'code')
     def on_change_with_code(self):
         if self.type and '_' in self.type:
@@ -587,6 +845,30 @@ class Identifier(sequence_ordered(), ModelSQL, ModelView):
                             code=self.code,
                             party=party))
 
+    @fields.depends(methods=['_notify_duplicate'])
+    def on_change_notify(self):
+        notifications = super().on_change_notify()
+        notifications.extend(self._notify_duplicate())
+        return notifications
+
+    @fields.depends('code', methods=['on_change_with_code'])
+    def _notify_duplicate(self):
+        cls = self.__class__
+        if self.code:
+            code = self.on_change_with_code()
+            others = cls.search([
+                    ('id', '!=', self.id),
+                    ('type', '!=', None),
+                    ('code', '=', code),
+                    ], limit=1)
+            if others:
+                other, = others
+                yield ('warning', gettext(
+                        'party.msg_party_identifier_duplicate',
+                        party=other.party.rec_name,
+                        type=other.type_string,
+                        code=other.code))
+
 
 class CheckVIESResult(ModelView):
     'Check VIES'
@@ -690,6 +972,7 @@ class Replace(Wizard):
         pool = Pool()
         Address = pool.get('party.address')
         ContactMechanism = pool.get('party.contact_mechanism')
+        Identifier = pool.get('party.identifier')
         transaction = Transaction()
 
         self.check_similarity()
@@ -702,6 +985,9 @@ class Replace(Wizard):
         ContactMechanism.write(list(source.contact_mechanisms), {
                 'active': False,
                 })
+        Identifier.write(list(source.identifiers), {
+                'active': False,
+                })
         source.replaced_by = destination
         source.active = False
         source.save()
@@ -709,9 +995,16 @@ class Replace(Wizard):
         cursor = transaction.connection.cursor()
         for model_name, field_name in self.fields_to_replace():
             Model = pool.get(model_name)
+            field = getattr(Model, field_name)
             table = Model.__table__()
             column = Column(table, field_name)
-            where = column == source.id
+            if field._type == 'reference':
+                source_value = str(source)
+                destination_value = str(destination)
+            else:
+                source_value = source.id
+                destination_value = destination.id
+            where = column == source_value
 
             if transaction.database.has_returning():
                 returning = [table.id]
@@ -722,7 +1015,7 @@ class Replace(Wizard):
 
             cursor.execute(*table.update(
                     [column],
-                    [destination.id],
+                    [destination_value],
                     where=where,
                     returning=returning))
 
@@ -737,6 +1030,7 @@ class Replace(Wizard):
         return [
             ('party.address', 'party'),
             ('party.contact_mechanism', 'party'),
+            ('party.identifier', 'party'),
             ]
 
 
@@ -749,7 +1043,6 @@ class ReplaceAsk(ModelView):
         domain=[
             ('id', '!=', Eval('source', -1)),
             ],
-        depends=['source'],
         help="The party that replaces.")
 
     @classmethod
@@ -780,21 +1073,9 @@ class Erase(Wizard):
         transaction = Transaction()
         cursor = transaction.connection.cursor()
 
-        def convert_from(table, tables):
-            right, condition = tables[None]
-            if table:
-                table = table.join(right, condition=condition)
-            else:
-                table = right
-            for k, sub_tables in tables.items():
-                if k is None:
-                    continue
-                table = convert_from(table, sub_tables)
-            return table
-
         resources = self.get_resources()
         parties = replacing = [self.ask.party]
-        with Transaction().set_context(active_test=False):
+        with inactive_records():
             while replacing:
                 replacing = Party.search([
                         ('replaced_by', 'in', list(map(int, replacing))),
@@ -811,7 +1092,7 @@ class Erase(Wizard):
                     Model.write(records, {'active': False})
 
                 tables, where = Model.search_domain(domain, active_test=False)
-                from_ = convert_from(None, tables)
+                from_ = convert_from(None, tables, type_='INNER')
                 table, _ = tables[None]
                 query = from_.select(table.id, where=where)
 
@@ -831,17 +1112,18 @@ class Erase(Wizard):
                                 where=table.id.in_(query)))
                 if resource:
                     for Resource in resources:
-                        model_tables = [Resource.__table__()]
+                        model_tables = [
+                            (Resource.__table__(), Resource.resource)]
                         if Resource._history:
-                            model_tables.append(Resource.__table_history__())
-                        for table in model_tables:
+                            model_tables.append(
+                                (Resource.__table_history__(),
+                                    Resource.resource))
+                        for (table, resource_field) in model_tables:
                             cursor.execute(*table.delete(
                                     where=table.resource.like(
                                         Model.__name__ + ',%')
-                                    & Model.id.sql_cast(
-                                        Substring(table.resource,
-                                            Position(',', table.resource)
-                                            + Literal(1))).in_(query)))
+                                    & resource_field.sql_id(
+                                        table.resource, Model).in_(query)))
         transaction.counter += 1
         return 'end'
 
